/*setup includes*/
#include <windows.h>
#include <stdio.h>
#include <commctrl.h>
#include "tasks.h"
#include "../ls-b24/lsapi/lsapi.h"
/*setup constants*/
const char AppName[] = "Tasks";
const int MAXTASKS = 16;
const int MAXICONS = 64;
const int MAXLEN = 256;
const int NUMTOKENS = 5;
/*setup variables*/
appIcon icons[MAXICONS];
BOOL autoArrange, forcePos, hidden, lbtnPressed, movingTask, noHints, noMove, taskMinimized, taskWrap, titles, titlesMinimized;
char defaultAppIcon[MAXLEN], fontName[MAXLEN], path[MAXLEN];
COLORREF backColor, darkColor, lightColor, textColor;
HBITMAP backBmp = NULL;
HCURSOR normalCursor, movingCursor;
HINSTANCE dll;
HMENU hMenu;
HWND parent, mainWin, deskTop, hints;
int direction, iconSize, iconPos, fontSize, numIcons, numTasks, prevTasks, screenHeight, screenWidth, taskPadding, taskSize, tasksX, tasksY, taskWrapNum, titlePos, wrapDirection;
POINT oldCursorPos;
taskType tasks[MAXTASKS];
UINT timer;
/*declare all the functions*/
BOOL IsAppWindow(HWND hWnd);
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
HICON GetIconFromWindow(HWND hWnd, BOOL bBigIcon);
int GetTaskByWnd(HWND hwnd);
int GetUnMovedNrTasks(void);
int GetAppIcon(int tasknr, char* target);
void CreateHints(HWND hWnd, char *txt, RECT *r);
void CreateTask(int i, HWND handle);
void DeleteTask(int i);
void GatherTasks(void);
void GetNewTaskPos(int i);
void GetTasks(void);
void PaintTask(int i);
void ReadConfig(void);
void ReArrangeTasks(BOOL all);
void RemoveHints(HWND hWnd);
void TaskPosition(int i, int tmpNumTasks);
void ToggleTasks(void);

/*and here's all the functions*/
int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd)
{
  return initModuleEx (ParentWnd, dllInst, wd->lsPath);
}

int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
  int i;
  RECT r;
  dll = dllInst;
  strcpy(path, szPath);
  parent = ParentWnd;
  GetClientRect(GetDesktopWindow(),&r);
  screenWidth = r.right;
  screenHeight = r.bottom;
  movingTask = false;
  deskTop = NULL;
  deskTop = FindWindow("DesktopBackgroundClass", NULL);
  if (!deskTop)
    deskTop = GetDesktopWindow();
  {
	WNDCLASS wc;
    memset(&wc,0,sizeof(wc));
	wc.lpfnWndProc = WndProc;
	wc.hInstance = dllInst;
	wc.lpszClassName = AppName;
	wc.style = CS_DBLCLKS;
	if (!RegisterClass(&wc)) 
	{
	  MessageBox(parent,"Error registering window class","Tasks",MB_OK);
	  return 1;
	}
  }
  normalCursor = GetCursor();
  if (!normalCursor)
	LoadCursor(NULL, IDC_ARROW);
  movingCursor = LoadCursor(NULL, IDC_SIZEALL);
  for (i = 0; i < MAXTASKS; i++)
  {
	tasks[i].hwnd = NULL;
	tasks[i].bmp = NULL;
	tasks[i].appHandle = NULL;
	tasks[i].appTitle = NULL;
	tasks[i].icon = NULL;
  }
  ReadConfig();
  mainWin = CreateWindowEx(WS_EX_TOOLWINDOW, AppName, "", WS_CHILD, 0, 0, 1, 1, deskTop, NULL, dll,	NULL);
  if (!mainWin) 
  {						   
    MessageBox(parent,"Error creating window","Tasks",MB_OK);
    return 1;
  }
  if (!noHints)
  {
	hints = CreateWindow(TOOLTIPS_CLASS,(LPSTR) NULL,TTS_ALWAYSTIP,	CW_USEDEFAULT,	CW_USEDEFAULT,	CW_USEDEFAULT,CW_USEDEFAULT,NULL,(HMENU) NULL, dll,	NULL);
	if (!hints) 
	{
	  MessageBox(parent, "Error creating hints window", "Tasks", MB_OK);
	  return 1;
	}
	SetWindowPos(hints, HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
  }
  SetWindowLong(mainWin, GWL_USERDATA, magicDWord);
  ShowWindow(mainWin, SW_HIDE);
  AddBangCommand("!GATHERTASKS", GatherTasks);
  AddBangCommand("!TOGGLETASKS", ToggleTasks);
  timer = SetTimer(mainWin, 0, 250, NULL);
  return 0;
}

void quitModule(HINSTANCE dllInst)
{
  int i;
  if (timer)
    KillTimer(NULL, timer);
  for(i = 0; i < numTasks; i++)
    DeleteTask(i);
  for(i = 0; i < numIcons; i++)
  {
	delete [] icons[numIcons].name;
	delete [] icons[numIcons].bitmap;
  }
  if (backBmp)
    DeleteObject(backBmp);
  DestroyWindow(mainWin);
  DestroyMenu(hMenu);
  if (!noHints) 
	DestroyWindow (hints);
  UnregisterClass(AppName,dllInst);
  return;
}

BOOL IsAppWindow(HWND hWnd)
{
	BOOL bResult = TRUE;
	LONG nStyle;
	HWND hOwner;
	HWND hParent;

	/*if it is a WS_CHILD or not WS_VISIBLE, fail it*/
	nStyle = GetWindowLong(hWnd, GWL_STYLE);	
	if((nStyle & WS_CHILD) || !(nStyle & WS_VISIBLE))
		bResult = FALSE;

	/*if the window is a WS_EX_TOOLWINDOW fail it*/
	nStyle = GetWindowLong(hWnd, GWL_EXSTYLE);
	if(nStyle & WS_EX_TOOLWINDOW)
		bResult = FALSE;

	/*If this has an owner, then only accept this window
	  if the partent is not accepted*/
	hOwner = GetWindow(hWnd, GW_OWNER);
	if(hOwner && bResult)
		bResult = !IsAppWindow(hOwner);

	/*if it has a parent, fail*/
	hParent = GetParent(hWnd);
	if(hParent)
		bResult = FALSE;

	if(GetWindowLong(hWnd, GWL_USERDATA) == 0x49474541)
		bResult = FALSE;

	return bResult;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
	case WM_TIMER:
	{
	  GetTasks();
      return 0;
	}
    case WM_QUERYENDSESSION:
	{
      return SendMessage(parent,message,wParam,lParam);
	}
	case WM_CREATE:
	  return 0;
	case WM_ERASEBKGND:
	  return 1; 
	case WM_PAINT:
	{
      int i = GetTaskByWnd(hwnd);
      if (i != -1)
		PaintTask(i);
	  return 0;
	}
	case WM_LBUTTONDOWN:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
	    GetCursorPos(&oldCursorPos);
	    SetCapture(tasks[n].hwnd);
		lbtnPressed = true;
		movingTask = false;
	  }
	  return 0;
	}
	case WM_LBUTTONUP:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
  	    POINT cursorPos;
	    GetCursorPos(&cursorPos);
  	    ReleaseCapture();
		lbtnPressed = false;
		if (GetCursor() != normalCursor)
	      SetCursor(normalCursor);
		if (abs(oldCursorPos.x - cursorPos.x) <= 5 && abs(oldCursorPos.y - cursorPos.y) <= 5 && tasks[n].hwnd == WindowFromPoint(cursorPos)) // click or canceled move
		{
		  if (!movingTask)
		  {
			if (IsIconic(tasks[n].appHandle))
			  ShowWindow(tasks[n].appHandle, SW_RESTORE);
			SendMessage(parent, LM_BRINGTOFRONT, 0, (long)tasks[n].appHandle);
			SetForegroundWindow(tasks[n].appHandle);
		  }
		  else /*canceled a move, force a redraw (I hate this, shouldnt be necesary)*/
		  {
			ShowWindow(tasks[n].hwnd, SW_HIDE);
		    ShowWindow(tasks[n].hwnd, SW_SHOWNORMAL);
		  }
		}
		else if (!noMove) /*move*/
		{
		  tasks[n].x -= oldCursorPos.x - cursorPos.x;
		  tasks[n].y -= oldCursorPos.y - cursorPos.y;
		  tasks[n].IsMoved = true;
		  SetWindowPos(tasks[n].hwnd, 0, tasks[n].x, tasks[n].y, taskSize, taskSize, 0); // SWP_NOZORDER|SWP_NOSIZE
		  ShowWindow(tasks[n].hwnd, SW_HIDE); /*stupid redraw again*/
		  ShowWindow(tasks[n].hwnd, SW_SHOWNORMAL);
		  if (autoArrange)
			ReArrangeTasks(false);
		}
	  }
	  return 0;
	}
	case WM_MOUSEMOVE:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (lbtnPressed && !noMove)
	  {
        POINT cursorPos;
		if (n != -1)
		{
	      GetCursorPos(&cursorPos);
	      if (abs(oldCursorPos.x - cursorPos.x) > 5 || abs(oldCursorPos.y - cursorPos.y) > 5 ||  tasks[n].hwnd != WindowFromPoint(cursorPos))
		  {
		    if (GetCursor() != movingCursor)
		      SetCursor(movingCursor);
			SetWindowPos(tasks[n].hwnd, 0, tasks[n].x - (oldCursorPos.x - cursorPos.x), tasks[n].y - (oldCursorPos.y - cursorPos.y), taskSize, taskSize, 0); //SWP_NOZORDER|SWP_NOSIZE
			movingTask = true;
		  }
		  else
		  {
		    if (GetCursor() != normalCursor)
		      SetCursor(normalCursor);
			SetWindowPos(tasks[n].hwnd, 0, tasks[n].x, tasks[n].y, taskSize, taskSize, 0); // SWP_NOZORDER|SWP_NOSIZE
		  }
		}
	  }
	  else
	  {
        if (!noHints) 
		{
		  MSG TooltipMessage;
		  TooltipMessage.hwnd=tasks[n].hwnd;
		  TooltipMessage.message=message;
		  TooltipMessage.wParam=wParam;
		  TooltipMessage.lParam=lParam;
		  SendMessage(hints,TTM_RELAYEVENT,0,(LPARAM)&TooltipMessage);
		  SetWindowPos(hints, HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
		}
	  }
	  return 0;
	}
	case WM_MOUSEACTIVATE:
		PostMessage(parent, 9185, 0, 0); /*Tells litestep it has been selected*/
		SendMessage(parent, 9183, (int)HIWORD(lParam), (int)LOWORD(lParam)); /*Close any popup open*/
		if (HIWORD(lParam) == WM_LBUTTONDOWN) /*te 10/30/98, Windows seems to be eating this?*/
			PostMessage(hwnd, WM_LBUTTONDOWN, 0, 0);
		return MA_ACTIVATE;
	case WM_ACTIVATE:
		SendMessage(parent, 9185, 0, 0); /*Tells litestep it has been selected*/
		if (LOWORD(wParam)) SetActiveWindow(parent);
		return 0;
	case WM_RBUTTONUP: /*Open popup menu*/
		SendMessage(parent, 9182, (int)HIWORD(lParam), (int)LOWORD(lParam));
		return 0;
	case WM_RBUTTONDOWN: /*Close popup menu*/
	case WM_MBUTTONDOWN:
		SendMessage(parent, 9183, (int)HIWORD(lParam), (int)LOWORD(lParam));
		return 0;
	case WM_KEYDOWN: 
	case WM_KEYUP:
	{	
	  PostMessage(parent,message,wParam,lParam);
	}
	return 0;
  }
  return DefWindowProc(hwnd,message,wParam,lParam);
}

HICON GetIconFromWindow(HWND hWnd, BOOL bBigIcon) /*taken from the re5ource popups, much credit to re5ource :)*/
{
   HICON hIcon = NULL;

   int nBigOrSmall = bBigIcon ? ICON_BIG : ICON_SMALL;
   int nBigOrSmall2 = bBigIcon ? GCL_HICON : GCL_HICONSM;

   SendMessageTimeout(hWnd, WM_GETICON, nBigOrSmall, 0, SMTO_ABORTIFHUNG, 1000, (PULONG)&hIcon);
   if(!hIcon) hIcon = (HICON)GetClassLong(hWnd, nBigOrSmall2);
   if(!hIcon) SendMessageTimeout(hWnd, WM_GETICON, nBigOrSmall, 1, SMTO_ABORTIFHUNG, 1000, (PULONG)&hIcon);
   if(!hIcon) hIcon = (HICON)GetClassLong(hWnd, nBigOrSmall2);
   if(!hIcon) SendMessageTimeout(hWnd, WM_QUERYDRAGICON, nBigOrSmall, 0, SMTO_ABORTIFHUNG, 1000, (PULONG)&hIcon);
   
   return hIcon;
}

int GetTaskByWnd(HWND hwnd)
{
  int i;
  for (i = 0;i < numTasks ; i++)
    if (tasks[i].hwnd == hwnd)
	  return i;
  return -1;
}

int GetUnMovedNrTasks(void)
{
  int result = 0;
  for (int i = 0; i < numTasks; i++)
    if (!tasks[i].IsMoved)
	  result++;
  return result;
}

int GetAppIcon(int tasknr, char* target)
{
  int i;
  for (i = 0; i < numIcons; i++)
  {
	if (strstr(tasks[tasknr].appTitle, icons[i].name))
	{
	  strcpy (target, icons[i].bitmap);
	  return 1;
	}
  }
  strcpy (target, defaultAppIcon);
  return 1;
}

void CreateHints(HWND hWnd, char *txt, RECT *r)
{ /*from the original ls code (shortcuts)*/
	TOOLINFO ti;    /*tool information*/
	if (noHints)
	{
		return;
	}

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.hinst = dll;
	ti.uId = 0;
	ti.lpszText = txt;
	ti.rect = *r;

	SendMessage(hints, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
  return;
}

void CreateTask(int i, HWND handle)
{
  RECT r;
  HWND hWnd = NULL;
  char frontImage[MAXLEN], title[MAXLEN];
  hWnd = CreateWindowEx(WS_EX_TOOLWINDOW, AppName, "", WS_CHILD, 0, 0,	taskSize, taskSize, deskTop, NULL, dll,	NULL);
  if (!hWnd) 
  {						   
    MessageBox(parent,"Error creating window","Tasks",MB_OK);
    return;
  }
  tasks[i].hwnd = hWnd;
  tasks[i].appHandle = handle;
  SetWindowLong(tasks[i].hwnd, GWL_USERDATA, magicDWord);
  GetWindowText(handle, (char *)&title, MAXLEN);
  tasks[i].appTitle = new char[strlen(title) + 1];
  strcpy(tasks[i].appTitle, title);
  GetNewTaskPos(i);
  tasks[i].IsIconic = IsIconic(handle);
  tasks[i].IsMoved  = false;
  GetAppIcon(i, frontImage);
  if (strcmp(".extract", frontImage) != 0)
	tasks[i].bmp = LoadLSImage(frontImage, NULL);
  else
	tasks[i].icon = GetIconFromWindow(tasks[i].appHandle, (iconSize > 31) ? true: false);
  SetWindowPos(tasks[i].hwnd, 0, tasks[i].x, tasks[i].y, taskSize, taskSize, 0);
  if (!hidden)
	ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
  else
	ShowWindow(tasks[i].hwnd, SW_HIDE);
  GetClientRect(tasks[i].hwnd, &r);
  CreateHints(tasks[i].hwnd, tasks[i].appTitle, &r);
  InvalidateRect(tasks[i].hwnd, &r, TRUE);
  return;
}

void DeleteTask(int i)
{
  RemoveHints(tasks[i].hwnd);
  if (tasks[i].hwnd)
    DestroyWindow(tasks[i].hwnd);
  if (tasks[i].bmp)
    DeleteObject(tasks[i].bmp);
  if (tasks[i].appTitle)
    delete [] tasks[i].appTitle;
  tasks[i].appHandle = NULL;
  tasks[i].hwnd = NULL;
  tasks[i].bmp = NULL;
  tasks[i].icon = NULL;  
  tasks[i].appTitle = NULL;
  return;
}

void GatherTasks(void)
{
  for (int i = 0; i < numTasks; i++)
	tasks[i].IsMoved = false;
  ReArrangeTasks(true);
  return;
}

void GetNewTaskPos(int i)
{
  TaskPosition(i, (autoArrange ? GetUnMovedNrTasks(): numTasks));
  return;
}

void TaskPosition(int i, int tmpNumTasks)
{
  int wrapLine = 0, resetLine = 0, checkLine = 0;
  wrapLine = tmpNumTasks / taskWrapNum;
  checkLine = tmpNumTasks - taskWrapNum * wrapLine;
  if (checkLine == 1)
	resetLine = 1;
  else
	resetLine = tmpNumTasks - taskWrapNum * wrapLine;
  if (direction == 1) /*up*/
  {
	if (taskWrap)
	{
	  if (tasksY >= 0)
	    tasks[i].y = tasksY - taskSize - resetLine * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + tasksY - taskSize - resetLine * (taskSize + taskPadding);
	  if (wrapDirection == 4) /*left*/
	  {
		if (tasksX >= 0)
		  tasks[i].x = tasksX - taskSize - wrapLine * (taskSize + taskPadding);
		else /*now accepts negative values*/
		  tasks[i].x = screenWidth + tasksX - taskSize - wrapLine * (taskSize + taskPadding);
	  }
	  else /*defaults to right, even if user inputs invalid up or down setting*/
	  {
	    if (tasksX >= 0)
	      tasks[i].x = tasksX + wrapLine * (taskSize + taskPadding);
	    else /*now accepts negative values*/
	      tasks[i].x = screenWidth + tasksX + wrapLine * (taskSize + taskPadding);
	  }
	}
	else
	{
	  if (tasksX >= 0)
	    tasks[i].x = tasksX;
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + tasksX;
	  if (tasksY >= 0)
	    tasks[i].y = tasksY - taskSize - tmpNumTasks * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + tasksY - taskSize - tmpNumTasks * (taskSize + taskPadding);
	}
  }
  else if (direction == 2) /*right*/
  {
    if (taskWrap)
	{
	  if (tasksX >= 0)
		tasks[i].x = tasksX + resetLine * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + tasksX + resetLine * (taskSize + taskPadding);
	  if (wrapDirection == 1) /*up*/
	  {
	    if (tasksY >= 0)
	      tasks[i].y = tasksY - taskSize - wrapLine * (taskSize + taskPadding);
	    else /*now accepts negative values*/
		  tasks[i].y = screenHeight + tasksY - taskSize - wrapLine * (taskSize + taskPadding);
	  }
	  else /*defaults to down, even if user inputs invalid left or right setting*/
	  {
	    if (tasksY >= 0)
	      tasks[i].y = tasksY + wrapLine * (taskSize + taskPadding);
	    else /*now accepts negative values*/
	      tasks[i].y = screenHeight + tasksY + wrapLine * (taskSize + taskPadding);
	  }
	}
	else
	{
	  if (tasksX >= 0)
		tasks[i].x = tasksX + tmpNumTasks * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + tasksX + tmpNumTasks * (taskSize + taskPadding);
	  if (tasksY >= 0)
		tasks[i].y = tasksY;
	  else /*now accepts negative values*/
		tasks[i].y = screenHeight + tasksY;
	}
  }
  else if (direction == 3) /*down*/
  {
    if (taskWrap)
	{
	  if (tasksY >= 0)
	    tasks[i].y = tasksY + resetLine * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + tasksY + resetLine * (taskSize + taskPadding);
	  if (wrapDirection == 4) /*left*/
	  {
		if (tasksX >= 0)
		  tasks[i].x = tasksX - taskSize - wrapLine * (taskSize + taskPadding);
		else /*now accepts negative values*/
		  tasks[i].x = screenWidth + tasksX - taskSize - wrapLine * (taskSize + taskPadding);
	  }
	  else /*defaults to right, even if user inputs invalid up or down setting*/
	  {
	    if (tasksX >= 0)
	      tasks[i].x = tasksX + wrapLine * (taskSize + taskPadding);
	    else /*now accepts negative values*/
	      tasks[i].x = screenWidth + tasksX + wrapLine * (taskSize + taskPadding);
	  }
	}
	else
	{
	  if (tasksX >= 0)
	    tasks[i].x = tasksX;
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + tasksX;
	  if (tasksY >= 0)
	    tasks[i].y = tasksY + tmpNumTasks * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + tasksY + tmpNumTasks * (taskSize + taskPadding);
	}
  }
  else if (direction == 4) /*left*/
  {
	if (taskWrap)
	{
	  if (tasksX >= 0)
	    tasks[i].x = tasksX - taskSize - resetLine * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + tasksX - taskSize - resetLine * (taskSize + taskPadding);
	  if (wrapDirection == 1) /*up*/
	  {
	    if (tasksY >= 0)
	      tasks[i].y = tasksY - taskSize - wrapLine * (taskSize + taskPadding);
	    else /*now accepts negative values*/
		  tasks[i].y = screenHeight + tasksY - taskSize - wrapLine * (taskSize + taskPadding);
	  }
	  else /*defaults to down, even if user inputs invalid left or right setting*/
	  {
	    if (tasksY >= 0)
	      tasks[i].y = tasksY + wrapLine * (taskSize + taskPadding);
	    else /*now accepts negative values*/
	      tasks[i].y = screenHeight + tasksY + wrapLine * (taskSize + taskPadding);
	  }
	}
	else
	{
	  if (tasksX >= 0)
	    tasks[i].x = tasksX - taskSize - tmpNumTasks * (taskSize + taskPadding);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + tasksX - taskSize - tmpNumTasks * (taskSize + taskPadding);
      if (tasksY >= 0)
        tasks[i].y = tasksY;
      else /*now accepts negative values*/
        tasks[i].y = screenHeight + tasksY;
	}
  }
  return;
}

void GetTasks(void)
{
  windowType * winList = NULL;
  char tmpTitle[MAXLEN];
  int m, n, maxWin;
  prevTasks = numTasks;
  numTasks = 0;		  
  winList = (windowType *)SendMessage(parent, LM_WINLIST, 0, 0);
  maxWin = (int) SendMessage(parent, LM_WINLIST, 1, 0);
  for (n = 0; n < maxWin && winList[n].Handle; n++)
  {
	if ((winList[n].Handle != NULL && numTasks < MAXTASKS) && (!taskMinimized || IsIconic(winList[n].Handle)) && IsAppWindow(winList[n].Handle))
	{
	  GetWindowText(winList[n].Handle, (char *)&tmpTitle, MAXLEN);
	  if (winList[n].Handle != tasks[numTasks].appHandle)
	  { 
		DeleteTask(numTasks);
	    CreateTask(numTasks, winList[n].Handle);
	    numTasks++;
	  }
	  else if (tasks[numTasks].IsIconic != IsIconic(winList[n].Handle))
	  {
		tasks[numTasks].IsIconic = IsIconic(winList[n].Handle);
		if (!hidden)
		{
		  ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
		  ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL);		
		  /*there must be a better way to force a redraw*/
		}
		numTasks++;
	  }
	  else if (strcmp(tmpTitle, tasks[numTasks].appTitle) != 0)
	  {
		  if (!noHints)
		  {
			TOOLINFO ti;
		    ti.cbSize = sizeof(TOOLINFO);
	  	    ti.hwnd = tasks[n].hwnd;
	        ti.uId = 0;
	        ti.lpszText = tmpTitle;
		    SendMessage(hints, TTM_UPDATETIPTEXT, 0, (LPARAM) (LPTOOLINFO) &ti);
		    delete [] tasks[numTasks].appTitle;
		    tasks[numTasks].appTitle = new char[strlen(tmpTitle) + 1];
		    strcpy(tasks[numTasks].appTitle, tmpTitle);
		    numTasks++;
		  }
	  }
	  else
	    numTasks++;
	}
  }
  for (m = numTasks; m < prevTasks; m++)
	DeleteTask(m);
  return;
}

void PaintTask(int i)
{
  PAINTSTRUCT ps;
  BITMAP bmInfo;
  HDC hdc = BeginPaint(tasks[i].hwnd,&ps);
  HDC src = CreateCompatibleDC(NULL);
  if (backBmp)
  {
    SelectObject(src, backBmp);
    TransparentBltLS(hdc, 0, 0, taskSize, taskSize, src, 0, 0, RGB(255, 0, 255));
  }
  if (tasks[i].bmp)
  {
    GetObject(tasks[i].bmp, sizeof(bmInfo), &bmInfo);
    if (bmInfo.bmWidth > taskSize)
      bmInfo.bmWidth = taskSize;
    if (bmInfo.bmHeight > taskSize) 
      bmInfo.bmHeight = taskSize;
    SelectObject(src, tasks[i].bmp);
	TransparentBltLS(hdc, (taskSize/2)-(bmInfo.bmWidth/2), (taskSize/2)-(bmInfo.bmHeight/2), bmInfo.bmWidth, bmInfo.bmHeight, src, 0, 0, RGB(255,0,255));
  }
  if (tasks[i].icon)
  {
	int posX, posY;
	posX = (taskSize / 2) - (iconSize / 2);
	if (titles || (titlesMinimized && tasks[i].IsIconic) || forcePos)
	  if (iconPos == 1)
		posY = 1;
	  else if (iconPos == 3)
		posY = taskSize - iconSize - 1;
	  else
		posY = (taskSize / 2) - (iconSize / 2);
	else
		posY = (taskSize / 2) - (iconSize / 2);
	if (iconSize == 32)
	  DrawIconEx(hdc, posX, posY, tasks[i].icon, 32, 32, 0, NULL, DI_NORMAL);
	else
      DrawIconEx(hdc, posX, posY, tasks[i].icon, 16, 16, 0, NULL, DI_NORMAL);
  }
  if (titles || (titlesMinimized && tasks[i].IsIconic))
  {
    char title[MAXLEN];
    HBRUSH backBrush;
    HFONT titleFont, oldFont;
    RECT r;
	HPEN pen, oldpen;
    GetClientRect(tasks[i].hwnd, &r);
	if (titlePos == 1)
	  r.top = 0;
	else
	  r.top = taskSize - fontSize - 3;
    r.bottom = r.top + fontSize + 2;
	if (backColor != 0x00FF00FF)
	{
      backBrush = CreateSolidBrush(backColor);
      FillRect(hdc, &r, backBrush);
	  /*Light Color Bevel (left/top)*/
	  pen = CreatePen(PS_SOLID, 1, lightColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.left, r.bottom - 1, NULL);
	  LineTo(hdc, r.left, r.top);
	  LineTo(hdc, r.right, r.top);
	  SelectObject(hdc, oldpen);
      DeleteObject(pen);
	  /*Dark Color Bevel (bottom/right)*/
      pen = CreatePen(PS_SOLID, 1, darkColor);
      oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.left, r.bottom, NULL);
      LineTo(hdc, r.right - 1, r.bottom);
	  LineTo(hdc, r.right - 1, r.top);
      SelectObject(hdc, oldpen);
      DeleteObject(pen);
	  DeleteObject(backBrush);
	}
	r.left += 2;
	r.right -= 2;
    int len = strlen(tasks[i].appTitle);
    titleFont = CreateFont(fontSize, 0, 0, 0, 0, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, fontName);
    oldFont = (HFONT) SelectObject(hdc, titleFont);
    SetBkMode(hdc, TRANSPARENT);
    SetTextColor(hdc, textColor);
    GetWindowText(tasks[i].appHandle, (char *)&title, MAXLEN);
    DrawText(hdc, title, len, &r, DT_VCENTER|DT_SINGLELINE|DT_END_ELLIPSIS);
    SelectObject(hdc, oldFont);
    DeleteObject(titleFont);
  }
  DeleteDC(src);
  EndPaint(tasks[i].hwnd,&ps);
  return;
}

void ReadConfig(void)
{
  FILE * f;
  unsigned int i;
  char * token[NUMTOKENS];
  char buffer[1024], extra[1024], tmpTBP[MAXLEN], tmpTD[MAXLEN], tmpTIP[MAXLEN], tmpTTP[MAXLEN], tmpTWD[MAXLEN];
  GetRCString("TasksAppIcon", defaultAppIcon, ".extract", MAXLEN);
  GetRCString("TasksBackPix", tmpTBP, ".none", MAXLEN);
	  if (strcmp(".none", tmpTBP))
		backBmp = LoadLSImage(tmpTBP, NULL);
	  else
		backBmp = NULL;
  GetRCString("TasksDirection", tmpTD, "right", MAXLEN);
		direction = 2;  /*only needed if user enters an invalid parameter*/
	  if (strcmp(tmpTD, "up") == 0)
		direction = 1;
	  else if (strcmp(tmpTD, "right") == 0)
		direction = 2;
	  else if (strcmp(tmpTD, "down") == 0)
		direction = 3;
	  else if (strcmp(tmpTD, "left") == 0)
		direction = 4;
  GetRCString("TasksIconPos", tmpTIP, "middle", MAXLEN);
		iconPos = 2; /*only needed if user enters an invalid parameter*/
	  if (strcmp(tmpTIP, "top") == 0)
		iconPos = 1;
	  else if (strcmp(tmpTIP, "middle") == 0)
		iconPos = 2;
	  else if (strcmp(tmpTIP, "bottom") == 0)
		iconPos = 3;
  GetRCString("TasksFont", fontName, "Arial", MAXLEN);
  GetRCString("TasksTitlePos", tmpTTP, "bottom", MAXLEN);
		titlePos = 2;  /*only needed if user enters an invalid parameter*/
	  if (strcmp(tmpTTP, "top") == 0)
		titlePos = 1;
	  else if (strcmp(tmpTTP, "bottom") == 0)
		titlePos = 2;
  GetRCString("TasksWrapDirection", tmpTWD, "down", MAXLEN);
		wrapDirection = 3;  /*only needed if user enters an invalid parameter*/
	  if (strcmp(tmpTWD, "up") == 0)
		wrapDirection = 1;
	  else if (strcmp(tmpTWD, "right") == 0)
		wrapDirection = 2;
	  else if (strcmp(tmpTWD, "down") == 0)
		wrapDirection = 3;
	  else if (strcmp(tmpTWD, "left") == 0)
		wrapDirection = 4;
  fontSize = GetRCInt("TasksFontSize", 12);
  iconSize = GetRCInt("TasksIconSize", 16);
	  iconSize = (iconSize > 31) ? 32 : 16;
  taskSize = GetRCInt("TasksSize", 32);
  taskPadding = GetRCInt("TasksPadding", 0);
  taskWrapNum = GetRCInt("TasksWrapCount", MAXTASKS);
	  if (taskWrapNum >= MAXTASKS || taskWrapNum < 2)
	  {
		taskWrapNum = MAXTASKS;
		taskWrap = false;
	  }
	  else
		taskWrap = true;
  tasksX = GetRCInt("TasksX", 0);
  tasksY = GetRCInt("TasksY", 0);
  autoArrange = GetRCBool("TasksAutoArrange", TRUE);
  forcePos = GetRCBool("TasksForceIconPos", TRUE);
  hidden = GetRCBool("TasksHidden", TRUE);
  noHints = GetRCBool("TasksNoHints", TRUE);
  noMove = GetRCBool("TasksNoMove", TRUE);
  taskMinimized = GetRCBool("TasksShowOnlyMinimized", TRUE);
  titles = GetRCBool("TasksTitlesAlways", TRUE);
  titlesMinimized = GetRCBool("TasksTitlesWhenMinimized", TRUE);
  backColor = GetRCColor("TasksBackColor", 0xFF00FF);
  lightColor = GetRCColor("TasksLightColor", 0xFFFFFF);
  darkColor = GetRCColor("TasksDarkColor", 0x000000);
  textColor = GetRCColor("TasksTextColor", 0xFFFFFF);

  numIcons = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < NUMTOKENS; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*Task", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, NUMTOKENS, extra);
	 if (count >= 3)
	 {
	   icons[numIcons].name = new char[strlen(token[1])+1];
	   icons[numIcons].bitmap = new char[strlen(token[2])+1];
	   strcpy(icons[numIcons].name, token[1]);
	   strcpy(icons[numIcons].bitmap, token[2]);
	   numIcons++;
	 }
  }
  LCClose(f);
  for (i = 0; i < NUMTOKENS; i++)
    delete [] token[i];
  return;
}

void ReArrangeTasks(BOOL all)
{
  int j = 0;
  for (int i = 0; i < numTasks; i++)
  {
	if (all || !tasks[i].IsMoved)
	{
	  TaskPosition(i, j);
      SetWindowPos(tasks[i].hwnd, 0, tasks[i].x, tasks[i].y, taskSize, taskSize, 0);
	  j++;
	}
  }
  all = false;
  return;
}

void RemoveHints(HWND hWnd)
{ /*from the original ls code (shortcuts)*/
	TOOLINFO ti;    /*tool information*/
	RECT r={0,0,0,0};

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.hinst = dll;
	ti.uId = 0;
	ti.lpszText = NULL;
	ti.rect = r;

	SendMessage(hints, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
  return;
}

void ToggleTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	if (hidden)
	  ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
	else
	  ShowWindow(tasks[i].hwnd, SW_HIDE);
  }
  hidden = !hidden;
  return;
}